#include <cstdlib>
#include <ctime>
#include <cstring>
#include <iostream>
#include <fstream>

#include "models.h"

using namespace std;





//******************************************************************************
//******************************************************************************
//
//  Parameter setting for the storage capacity
//
//******************************************************************************
//******************************************************************************

//The maximum capacity (maximum number of characters allowed) 
//		for the storing the vocabulary set
#define VocabularyStorageLimit 20000

//The maximum capacity (maximum number of characters allowed) 
//		for storing the corrupted words during spelling recognition
#define NoisyWordsStorageLimit 15000






//******************************************************************************
//******************************************************************************
//
//  Parameter setting for the keyboard model
//
//******************************************************************************
//******************************************************************************

double prExact = 0.8;
//The probability that you want to type a character and you do successfully make it
//In the setting of the original keyboard model in the project,
//we make it 0.8, but you can try different constants to see the effects.

double prNearby = 0.2;
//The sum of probabilities that you want to type a character but end in touching 
//a different character.
//In the setting of the original keyboard model in the project,
//we make it 0.2, but you can try different constants to see the effects.


int LikelihoodDistancePower = 1;
//The likelihood you want to type a character but end in touching a different character
//is proportion to the inverse of 
//(the distance between them) to the (LikelihoodDistancePower)th power
//In the setting of the original keyboard model in the project,
//we make it 1, but you can try different constants to see the effects.








//******************************************************************************
//******************************************************************************
//
//  Parameter setting for the Spelling model
//
//******************************************************************************
//******************************************************************************
double prRepeat = 0.1;
//The probability of repeating the current cognitive state again as the next state

double prMoveOn = 0.9;
//The probability of moving from the current cognitive state to other states
//	as the next state

int InitialLikelihoodDistancePower = 4;
//The likelihood of some character in a word as the initial cognitive state
//is proportion to the inverse of 
//(the position of the character in the word) to the 
//(InitialLikelihoodDistancePower)th power.
//In the setting of the original spelling model in the project,
//we make it 2, but you can try different constants to see the effects.

int TransitionLikelihoodDistancePower = 5;
//The likelihood of moving from the cognitive state of typing some character in a word 
//to the next cognitive state is proportion to the inverse of the distance between
//(the current state to the next state) to the 
//(TransitionLikelihoodDistancePower)th power.
//In the setting of the original spelling model in the project,
//we make it 2, but you can try different constants to see the effects.





//******************************************************************************
//******************************************************************************
//
//  Trace flag and the alphabet table
//
//******************************************************************************
//******************************************************************************
bool traceON = true;   // output tracing messages

char asciiTable[] = "abcdefghijklmnopqrstuvwxyz";
//The table storing the 26 letters of the English alphabet






//******************************************************************************
//******************************************************************************
//
//  Implementation of functions
//
//******************************************************************************
//******************************************************************************


/************************************************************************/
//Simulate the combination of the spelling model and the keyboard model:
//Given a word stored in the word array, simulate what may actually
//be typed and store the result in the output array.
//maxOutput specifies the capacity limit of the output array
/************************************************************************/
void typeOneWord(char word[], char output[], int maxOutput)
{ double temp;  int i;

  int numChars;
  numChars = strlen(word);

  double * initialPrTable = new double[numChars];  
  int sizeOfSpace = numChars;
  getPrTableForPossibleInitialStates(initialPrTable,sizeOfSpace);

  int initialState, currentState;
  initialState = take1SampleFrom1PrSpace(initialPrTable,sizeOfSpace);
  currentState = initialState;

  double * transitionPrTable = new double[numChars+1];
  bool finalStateReached = false;
  int outputIndex = 0;
  bool outputOutOfSpace = false;
  //Repeatedly do the following:
  //generate one character based on the current state, 
  //calculate the transition probabilities to the possible next states, and
  //then move on to the next state.
  while (!finalStateReached && !outputOutOfSpace)
  {	
	output[outputIndex++] = typeOneChar( word[currentState] );
	if (outputIndex > maxOutput -2)
	{ outputOutOfSpace = true;
	  cout << "The output buffer is out of space." << endl;
	}

	//	Transition to a different cognitive state to the right of the current state
	/*******************************************************************/
	int nextState;
	getPrTableForPossibleNextStates(transitionPrTable, numChars+1, currentState);
	nextState = take1SampleFrom1PrSpace(transitionPrTable,numChars+1);
	int FinalState = numChars;
	if (nextState == FinalState)
		finalStateReached = true;

	if (traceON)
	{
		cout << "Current state:" << word[currentState] << ',' << currentState;
		if (finalStateReached)
			cout << " The next state: The final state" << endl;
		else
			cout << " The next state: " << word[nextState]
				 << ',' << nextState << endl;
	}

	currentState = nextState;

  }//end of while loop of state transitions

  output[outputIndex] = '\0'; //Terminate the output string with '\0'
  delete [] initialPrTable;
  delete [] transitionPrTable;

}// end of the function
/*******************************************************************/



/************************************************************************/
//Simulate the keyboard model:
//Given the charToTye, simulate what character may actually
//be typed and return it as the result.
/************************************************************************/
char typeOneChar(char charToType)
{ double temp, alpha;
  int distanceTable[26];
  double likelihoodTable[26];
  double prTable[26];
  double prIntervals[26];
  int indexOfCharToType, indexCharGenerated;
  int i,distance;

  //Calculate the distance of other characters from the character to type
  charToType = tolower(charToType);
  if (charToType<'a' || charToType >'z')
	  cout << "The simulated keyboard does only support the English alphbet" << endl;
  indexOfCharToType = charToType - 'a';
  for (i=0; i<26; i++)
  { distance = charToType - asciiTable[i];
    if (distance < 0)
		distance = distance * (-1);
    if (distance > 13)
		distance = 26 - distance;
	distanceTable[i] = distance;
  }

  /*
  for (i=0; i<26; i++)
  { cout << asciiTable[i] << ":" << distanceTable[i] 
         << " characters away from " << charToType << endl;
  }
  */

  //Calculate the likelihood indicators of the characters
  for (i=0; i<26; i++)
  { int distanceInPower, j;
    distanceInPower = 1;
	for (j= 0; j<LikelihoodDistancePower; j++)
		distanceInPower = distanceInPower * distanceTable[i];
	if (i != indexOfCharToType)
		likelihoodTable[i] = 1.0 / distanceInPower;
    /* if (i == indexOfCharToType) 
		  it is not defined.
    */
  }

   //Calculate the scaling factor alpha for the likelihood indicators of the keyboard model
  double sumOfLikelihood=0;
  for (i=0; i<26; i++)
    if (i != indexOfCharToType)
	  sumOfLikelihood = sumOfLikelihood + likelihoodTable[i];
  alpha = prNearby / sumOfLikelihood;
  // prNearby is 0.2, the probability that you want to type a particular but has a typo.
  // And aplpha times sumOfLikelihood should equal prNearby.
  // cout << " alpha in the keyboard model is " << alpha << endl;


  //Calculate the probability of the characters
  for (i=0; i<26; i++)
  { double prTouchChar;
    if (i == indexOfCharToType)
		prTable[i]= prExact;
	else
		prTable[i]= alpha * likelihoodTable[i];
  }

  double prSum=0;
  for (i=0; i<26; i++)
    prSum += prTable[i];
  if (prSum > 1.001 || prSum < 0.999)
	  cout << "["<< charToType << "]:Wrong:"
	       << "Sum of all probabilities is not 1.0, but" 
		   << prSum  << endl;

  indexCharGenerated = take1SampleFrom1PrSpace(prTable,26);
  return asciiTable[indexCharGenerated];

}

/************************************************************************/
//Calculate and return the probability of charGenerated actually typed
//given charOfTheState as the underlying cognitive state. 
/************************************************************************/
double prCharGivenCharOfState(char charGenerated, char charOfTheState)
{ double temp, alpha;
  int distanceTable[26];
  double likelihoodTable[26];
  int indexOfCharToType, indexOfCharGenerated;
  int i,distance;

  //Calculate the distance of other characters from the character to type
  char charToType = charOfTheState;
  charGenerated = tolower(charGenerated);
  if  (charToType<'a' || charToType >'z'|| charGenerated<'a' || charGenerated >'z')
  {cout << "The simulated keyboard does not support things other than the English alphbet."
        << endl;
   return 0;
  }

  indexOfCharToType = charToType - 'a';
  for (i=0; i<26; i++)
  { distance = charToType - asciiTable[i];
    if (distance < 0)
		distance = distance * (-1);
    if (distance > 13)
		distance = 26 - distance;
	distanceTable[i] = distance;
  }

  /*
  for (i=0; i<26; i++)
  { cout << asciiTable[i] << ":" << distanceTable[i] 
         << " characters away from " << charToType << endl;
  }
  */

  //Calculate the likelihood indicators of the characters
  for (i=0; i<26; i++)
  { int distanceInPower, j;
    distanceInPower = 1;
	for (j= 0; j<LikelihoodDistancePower; j++)
		distanceInPower = distanceInPower * distanceTable[i];
	if (i != indexOfCharToType)
		likelihoodTable[i] = 1.0 / distanceInPower;
    /* if (i == indexOfCharToType) 
		  it is not defined.
    */
  }

   //Calculate the scaling factor alpha for the likelihood indicators of the keyboard model
  double sumOfLikelihood=0;
  for (i=0; i<26; i++)
    if (i != indexOfCharToType)
	  sumOfLikelihood = sumOfLikelihood + likelihoodTable[i];
  alpha = prNearby / sumOfLikelihood;
  // prNearby is 0.2, the probability that you want to type a particular but has a typo.
  // And aplpha times sumOfLikelihood should equal prNearby.
  // cout << " alpha in the keyboard model is " << alpha << endl;


  //Calculate the probability of the characters
  indexOfCharGenerated = charGenerated - 'a';
  if (indexOfCharGenerated == indexOfCharToType)
		return prExact;
  else
		return alpha * likelihoodTable[indexOfCharGenerated];

}

/************************************************************************/
//Given the probabilities (of sizeOfTable elements) stored in prTable,
//	try to randomly take a sample out of sizeOfTable elements
//	according to the probabilities (of sizeOfTable elements;
//Return the index of the non-deterministically sampled element.
/************************************************************************/
int take1SampleFrom1PrSpace(double prTable[], int sizeOfTable)
{ int i;
  double prSum=0;
  for (i=0; i<sizeOfTable; i++)
    prSum += prTable[i];
  if (prSum < 0.999 || prSum > 1.001)
		cout << "Something is wrong with a random sampleing" << endl
		<< "The sum of all probabilities in the table is " << prSum  << endl;

   //Calculate the probability intervals of the characters
  double prAccumulated = 0;
  double * prIntervals;
  prIntervals = new double[sizeOfTable];
  for (i=0; i<sizeOfTable; i++)
  { prAccumulated = prAccumulated + prTable[i];
    prIntervals[i] = prAccumulated;
  }

  /*
  cout << endl;
  for (i=0; i<sizeOfTable; i++)
  { if (i>0) 
		cout << "state " << i << " lower bounded by " << prIntervals[i-1]<< endl;
    cout << "state " << i << " upper bounded by " << prIntervals[i]<< endl;
  }
  */

  // Generate a random number in [0,1]
  i = rand()% 1001; 
  double temp = i / 1001.0;
  //cout << "The random number pair generated is " << i << endl << temp << endl;

  bool sampleTaken = false;
  for (i=0; i<sizeOfTable && !sampleTaken; i++)
	if (temp <= prIntervals[i])
	{   delete [] prIntervals;
		return i;
	}
  return sizeOfTable-1;
}




/************************************************************************/
//Calculate for each actual cognitive state in a word of sizeOfTable characters,
//	the probability of that cognitive state being the actual first cognitive state
//	when the user types the word.
//Store these prbabilities in the prTable array.
/************************************************************************/
void getPrTableForPossibleInitialStates(double prTable[], int sizeOfTable)
{
  int i;
  int initialState, state;
  double * initialLikelihoodTable;
  initialLikelihoodTable = new double[sizeOfTable]; 
  //One slot for each regular state

  for (int i=0; i<sizeOfTable; i++)
  { int positionInPower, j;
    positionInPower = 1;
	for (j= 0; j<InitialLikelihoodDistancePower; j++)
		positionInPower = positionInPower * (i+1);
  	initialLikelihoodTable[i]= 1.0/positionInPower;
  }

  // Calculate the scaling factor alpha for the likelihood indicators of 
  // the possible initial states of the spelling model for this particular word.
  double sumOfLikelihood=0;
  for (i=0; i<sizeOfTable; i++)
  	  sumOfLikelihood = sumOfLikelihood + initialLikelihoodTable[i];
  double beta = 1.0 / sumOfLikelihood;
  // One of the states must be the initial state.
  // So aplpha times sumOfLikelihood should equal 1.0.
  // cout << " beta in this spelling model is " << beta << endl;

  //For each character in the word, calculate the probability of being the initial state
  for (i=0; i<sizeOfTable; i++)
  { 
		prTable[i]= beta * initialLikelihoodTable[i];
  }

  delete [] initialLikelihoodTable;
}






/************************************************************************/
//Calculate for each actual cognitive state in a word of sizeOfTable characters,
//	the probability of that cognitive state being the next cognitive state
//	given currentState as the current state.
//Store these prbabilities in the transitionPrTable[] array.
/************************************************************************/
void getPrTableForPossibleNextStates
(double transitionPrTable[], int sizeOfTable, int currentState)
{
  int i;
  double * transitionLikelihoodTable = new double[sizeOfTable]; 
  //One slot for each regular state plus the final state

	for (int i=0; i<sizeOfTable; i++)
		{ int distanceInPower, j;
		  distanceInPower = 1;
		  if (i<=currentState)
			//Never back to an earlier state if move on to a different state
			transitionLikelihoodTable[i]=0; 
		  else
		   {int nextStateDistance;
		    nextStateDistance = i - currentState;
			for (j= 0; j<TransitionLikelihoodDistancePower; j++)
				distanceInPower = distanceInPower * nextStateDistance;
			transitionLikelihoodTable[i]= 1.0/distanceInPower;
		   }
		}

	// Calculate the scaling factor alpha for the likelihood indicators of 
	// the possible initial states of the spelling model for this particular word.
	double sumOfLikelihood=0;
	for (i=0; i<sizeOfTable; i++)
  			sumOfLikelihood = sumOfLikelihood + transitionLikelihoodTable[i];
	double gamma = prMoveOn / sumOfLikelihood;
	// One of the states must be the next state.
	// So gamma times sumOfLikelihood should equal 0.9.
	// cout << " gamma in this spelling model is " << gamma << " " << sumOfLikelihood << endl;

	//Never back to an earlier state to the left of the current state
	for (i=0; i<currentState; i++)
		{ 
			transitionPrTable[i]= 0;
		}

	//The prbaibilty of repeating the same state
	transitionPrTable[currentState]=prRepeat;

	//Calculate the probability of moving onto a state to the right of current state
	for (i=currentState+1; i<sizeOfTable; i++)
		{ 
			transitionPrTable[i]= gamma * transitionLikelihoodTable[i];
			//cout << "*********" << transitionPrTable[i] << endl;
		}

  delete [] transitionLikelihoodTable;
}




/************************************************************************/
//The probability of getting a series of characters stored in charSeries 
//	(terminated with an additional NULL character in the charSeries array)
//when you want to type a word stored in the word array
//	(terminated with an additional NULL character in the word array).
/************************************************************************/
double prOf1CharSeriesWhenTyping1Word(char charSeries[], char word[])
{int i;
 int numCharsInWord = strlen(word);
 int numCharsInCharSeries = strlen(charSeries);
 int numOfRegularStages = numCharsInCharSeries;
 int numStatesExceptTheFinalState = numCharsInWord;
 int numStatesIncludeTheFinalState = numCharsInWord+1;

 double * prTableThisStage = new double[numStatesExceptTheFinalState];
 double * prTableNextStage = new double[numStatesExceptTheFinalState];

 getPrTableForPossibleInitialStates(prTableThisStage, numStatesExceptTheFinalState);
 int state, nextState, stageFinished, indexOfCharObserved;
 stageFinished = 0;


 indexOfCharObserved = 0;

 for (state=0; state<numStatesExceptTheFinalState; state++)
	prTableThisStage[state] *=	
				prCharGivenCharOfState(charSeries[indexOfCharObserved],word[state]);
 

 double * prTableForPossibleNextStates = new double[numStatesIncludeTheFinalState];
 for (stageFinished = 1; stageFinished<numOfRegularStages; stageFinished++)
 {	
	/*
	cout << "Result of Stage " << stageFinished - 1 << " :" << endl;
	for (state=0; state<numStatesExceptTheFinalState; state++)
		cout << "\t[" << state << "]:\t" << prTableThisStage[state];
	cout << endl;
	*/
	 
	for (state = 0; state < numStatesExceptTheFinalState; state++)
		prTableNextStage[state] = 0;

	for (state = 0; state < numStatesExceptTheFinalState; state++)
	{
		getPrTableForPossibleNextStates
			(prTableForPossibleNextStates, numStatesIncludeTheFinalState, state);

		for (nextState = 0; nextState < numStatesExceptTheFinalState; nextState++)
			prTableNextStage[nextState] += 
				( prTableThisStage[state] * prTableForPossibleNextStates[nextState]);
	}

	indexOfCharObserved = stageFinished;
	for (state=0; state<numStatesExceptTheFinalState; state++)
	prTableNextStage[state] *=	
			prCharGivenCharOfState(charSeries[indexOfCharObserved],word[state]);

	double * temp;
	temp = prTableThisStage;
	prTableThisStage = prTableNextStage;
	prTableNextStage = temp;
 }
 
 /*
 cout << "Result of Stage " << stageFinished - 1 << " :" << endl;
 for (state=0; state<numStatesExceptTheFinalState; state++)
	cout << "\t[" << state << "]:\t" << prTableThisStage[state];
 cout << endl;
 */

 double prTransitionToFinalState=0;
 int finalState= numCharsInWord;
 for (state = 0; state < numStatesExceptTheFinalState; state++)
	{	
		getPrTableForPossibleNextStates
			(prTableForPossibleNextStates, numStatesIncludeTheFinalState, state);

		prTransitionToFinalState += 
				(prTableThisStage[state] * prTableForPossibleNextStates[finalState]);
	}

 return prTransitionToFinalState;

}



/************************************************************************/
/************************************************************************/
void typeRandomWordsFromVocabulary
(int numWordsToType, char * corruptedMessageFile, char * realMessageFile, 
 char * vocabularyFile, bool trace)
//Randomly select numWordsToType words (with repition allowed) from the vocabularyFile
//Store these numWordsToType randomly selected words into realMessageFile.
//Type these numWordsToType words one by one into corruptedMessageFile
//according to the acoutic model.
//If trace is true, then additional information is provided on 
//how the noisy words are generated.
{
	int i;
	ifstream vocabularyFileHandle;
	char vocabularyBuffer[VocabularyStorageLimit];
	int numWordsInVocabulary;
	numWordsInVocabulary = 
		readVocabularyFileIntoCharBuffer(vocabularyBuffer, VocabularyStorageLimit-1, vocabularyFile);

	char * * vocabularyPointersList;
	vocabularyPointersList = new char * [numWordsInVocabulary];
	char * ptrIntoBuffer = vocabularyBuffer;
	for (i=0; i<numWordsInVocabulary; i++)
	{	vocabularyPointersList[i] = ptrIntoBuffer;
		if (i != numWordsInVocabulary-1)
			while ( *ptrIntoBuffer++ != '\0');
	}

	if (trace)
	{
		cout << "List of words in the vocabulary file:" << vocabularyFile<< endl;
		for (int i=0; i<numWordsInVocabulary; i++)
			cout << vocabularyPointersList[i] << endl;
	}

	ofstream realMessageFileHandle;
	realMessageFileHandle.open(realMessageFile);
	if ( realMessageFileHandle.fail() )
	{	cout << "cannot open the realMessageFile file:" << realMessageFile << endl;
		return;
	}

	ofstream corruptedMessageFileHandle;
	corruptedMessageFileHandle.open(corruptedMessageFile);
	if ( vocabularyFileHandle.fail() )
	{	cout << "cannot open the corruptedMessageFile file:" << corruptedMessageFile << endl;
		return;
	}

    int random;
	char wordBuffer[100];
	for (i=0; i<numWordsToType; i++)
	{	random = rand() % numWordsInVocabulary;
		cout << endl
			<< "Simulate the typing of this word:" << vocabularyPointersList[random] << endl;
		typeOneWord(vocabularyPointersList[random] , wordBuffer, 99);
		cout << endl;

		realMessageFileHandle << vocabularyPointersList[random] << endl;
		corruptedMessageFileHandle << wordBuffer << endl;

		if (trace)
			corruptedMessageFileHandle <<"Try to type the word:" 
				<< vocabularyPointersList[random]
				<< "\t\t and/but we get the word\t" 
				<< wordBuffer << endl;
	}
	delete [] vocabularyPointersList;
	corruptedMessageFileHandle.close();
	realMessageFileHandle.close();
}



/************************************************************************/
/************************************************************************/
void typeOneArticle
(char * corruptedMessageFile, char * sourceArticle, bool trace)
//Read the words from the file: sourceArticle one by one and
//	simulate how each word may be typed according to the spelling model
//	and the keyboard model and store the resulting words in the file corruptedMessageFile.
{
	//Read abd store the words in the source article into the vocabulary buffer 
	int i;
	ifstream vocabularyFileHandle;
	char vocabularyBuffer[VocabularyStorageLimit];
	int numWordsInVocabulary;
	numWordsInVocabulary = 
		readVocabularyFileIntoCharBuffer(vocabularyBuffer, VocabularyStorageLimit-1, sourceArticle);

	char * * vocabularyPointersList;
	vocabularyPointersList = new char * [numWordsInVocabulary];
	char * ptrIntoBuffer = vocabularyBuffer;
	for (i=0; i<numWordsInVocabulary; i++)
	{	vocabularyPointersList[i] = ptrIntoBuffer;
		if (i != numWordsInVocabulary-1)
			while ( *ptrIntoBuffer++ != '\0');
	}

	if (trace)
	{
		cout << "List of words in the source article:" << sourceArticle<< endl;
		for (int i=0; i<numWordsInVocabulary; i++)
			cout << vocabularyPointersList[i] << endl;
	}


	ofstream corruptedMessageFileHandle;
	corruptedMessageFileHandle.open(corruptedMessageFile);
	if ( vocabularyFileHandle.fail() )
	{	cout << "cannot open the corruptedMessageFile file:" << corruptedMessageFile << endl;
		return;
	}


	char wordBuffer[100];
	for (i=0; i<numWordsInVocabulary; i++)
	{	
		cout << endl
			<< "Simulate the typing of this word:" << vocabularyPointersList[i] << endl;
		typeOneWord(vocabularyPointersList[i] , wordBuffer, 99);
		cout << endl;

		corruptedMessageFileHandle << wordBuffer << endl;

		if (trace)
			corruptedMessageFileHandle <<"Try to type the word:" 
				<< vocabularyPointersList[i]
				<< "\t\t and/but we get the word\t" 
				<< wordBuffer << endl;
	}
	delete [] vocabularyPointersList;
	corruptedMessageFileHandle.close();
}




/************************************************************************/
/************************************************************************/
void recognize1NoisyWordFromVocabulary
(char * noisyWordToRecognize, 
 char * * vocabularyPointersList, int sizeOfVocabulary,
 char * outputBuffer, int bufferSize, int numOfLikelyWordsToSuggest,
 bool trace)
//Recongnize the noisyWrodToRecognize, suggest up to numOfLikelyWordsToSuggest 
//most likely words out of the sizeOfVocabulary words inthe vocabulary
//pointed to by pointers in vocabularyPointersList, and write these words to the
//outputBuffer of size bufferSize.
//If the outputBuffer is not big enough to hold all the numOfLikelyWordsToSuggest
//words, return false; otherwise return true.
{
	int i;
	int * candidateWordsIndices;
	candidateWordsIndices = new int[numOfLikelyWordsToSuggest];
	for (int i=0; i<numOfLikelyWordsToSuggest; i++)
		candidateWordsIndices[i] = -1;

	double * candidateWordsPrTable;
	candidateWordsPrTable = new double[numOfLikelyWordsToSuggest];
	for (int i=0; i<numOfLikelyWordsToSuggest; i++)
		candidateWordsPrTable[i] = 0;

	double currentMinimalPrInTable=0;
	int indexOfCurrentMinimalPrInTable = 0;
	for (int wordIndex=0; wordIndex<sizeOfVocabulary; wordIndex++)
	{	
		double likelihoodOfThisWord;
		likelihoodOfThisWord = 
			prOf1CharSeriesWhenTyping1Word
			(noisyWordToRecognize, vocabularyPointersList[wordIndex]);
		cout << "Pr(" << noisyWordToRecognize << '|' 
			 <<  vocabularyPointersList[wordIndex] << ")="
			 << likelihoodOfThisWord << endl;
	
		
		if (likelihoodOfThisWord > currentMinimalPrInTable)
		{	//Find a good candidate word
			/*
			cout << "[" << vocabularyPointersList[wordIndex] << "] "  
				 << "has a higher probability than " <<currentMinimalPrInTable 
				 << ", and now it is stored in "
				 << indexOfCurrentMinimalPrInTable 
				 << "th entry of the most likely candidate list." << endl ;
			*/

			candidateWordsIndices[indexOfCurrentMinimalPrInTable] = wordIndex;
			candidateWordsPrTable[indexOfCurrentMinimalPrInTable] = likelihoodOfThisWord;

			//Update currentMinimalPrInTable and indexOfCurrentMinimalPrInTable
			currentMinimalPrInTable = candidateWordsPrTable[0];
			indexOfCurrentMinimalPrInTable = 0;
			for (i = 0; i< numOfLikelyWordsToSuggest; i++)
					if ( candidateWordsPrTable[i] < currentMinimalPrInTable )
					{	currentMinimalPrInTable = candidateWordsPrTable[i];
						indexOfCurrentMinimalPrInTable = i;
					}
		}	
		
	}

	//Use selection sort to sort the cadidate words in order
	 for (int stageFinished = 0; stageFinished< numOfLikelyWordsToSuggest; stageFinished++)
		{	int indexNextMaxPrInTable;
			double nextMaxPrInTable;
			indexNextMaxPrInTable = stageFinished;
			nextMaxPrInTable = candidateWordsPrTable[stageFinished];
			for (i = stageFinished+1; i< numOfLikelyWordsToSuggest; i++)
					if ( candidateWordsPrTable[i] > nextMaxPrInTable )
					{	nextMaxPrInTable = candidateWordsPrTable[i];
						indexNextMaxPrInTable = i;
					}

			double tempPr;
			int tempIndex;
			tempIndex = candidateWordsIndices[stageFinished];
			tempPr = candidateWordsPrTable[stageFinished];

			candidateWordsIndices[stageFinished]=
							candidateWordsIndices[indexNextMaxPrInTable];
			candidateWordsPrTable[stageFinished]=
							candidateWordsPrTable[indexNextMaxPrInTable];

			candidateWordsIndices[indexNextMaxPrInTable]= tempIndex;
			candidateWordsPrTable[indexNextMaxPrInTable]= tempPr;
	}

	int ouputIndex = 0;
	char *sourcePtr;
	char *targetPtr = outputBuffer;
	if (trace) //Send the output to the screen if we want visual trace
		for (i = 0; i< numOfLikelyWordsToSuggest; i++)
		{	cout << i <<"th possible candidate word for [" << noisyWordToRecognize << "] is:"
				 << vocabularyPointersList[ candidateWordsIndices[i] ] << endl
				 << "The probability of it is:" 
				 << candidateWordsPrTable[ i ] 
				 << endl << endl << endl;
		}
	else
		for (i = 0; i< numOfLikelyWordsToSuggest; i++)
		{	sourcePtr = vocabularyPointersList[ candidateWordsIndices[i] ];
			while ( (targetPtr< outputBuffer + bufferSize-1) && *sourcePtr != '\0')
				*targetPtr++ = *sourcePtr++;
			*targetPtr++ = '\t';
		}
	*targetPtr++ = '\0';
	
	delete [] candidateWordsIndices;
	delete [] candidateWordsPrTable;

}



/************************************************************************/
/************************************************************************/
int readVocabularyFileIntoCharBuffer
(char * vocabularyBuffer, int sizeOfVocabularyBuffer,
 char * vocabularyFile)
//Read in words from vocabularyFile and store each word as a contiguous series
//of characters ended with an additional NULL character in the vocabularyBuffer. 
//The vocabularyBuffer has a capcility of sizeOfVocabularyBuffer.
//We allow up to 99 characters per word.
//Return the total number of words read from vocabularyFile.
{
	int i;
	ifstream vocabularyFileHandle;
	vocabularyFileHandle.open(vocabularyFile);
	if ( vocabularyFileHandle.fail() )
	{	cout << "cannot open the vocabulary file:" << vocabularyFile << endl;
		return 0;
	}

	int totalWords = 0;
	int vocabularyBufferIndex = 0;
	char wordBuffer[100]; 
	//at most 99 character per word plus the terminating NULL character
	while ( !vocabularyFileHandle.eof() )
	{	int charIndex;
		vocabularyFileHandle >> wordBuffer;
		if ( (vocabularyBufferIndex + strlen( wordBuffer )) >= sizeOfVocabularyBuffer-1)
		{	cout << "Not enough buffer space to read the entire vocabulary" << endl;
			vocabularyFileHandle.close();
			return totalWords;
		}
		if ( strlen( wordBuffer ) > 0 )  
			//A real word to be copied into the buffer
		{	charIndex = 0;
			while (wordBuffer[charIndex] != '\0')
				vocabularyBuffer[vocabularyBufferIndex++] = wordBuffer[charIndex++];
			vocabularyBuffer[vocabularyBufferIndex++] = '\0';
			totalWords++;
		}
	}
	vocabularyFileHandle.close();
	return totalWords;
}


/************************************************************************/
/************************************************************************/
int getWordsPointersListFromCharBuffer
(char * * vocabularyPointersList, int sizeOfVocabularyPointersList,
 char * vocabularyBuffer, int sizeOfVocabularyBuffer)
//Parse the vocabularyBuffer of size sizeOfVocabularyBuffer, and
//store up to sizeOfVocabularyPointersList character pointers pointing to 
//up to the first sizeOfVocabularyPointersList words stored in vocabularyBuffer.
{
	int i;
	int charCount = 0;
	char * ptrIntoBuffer = vocabularyBuffer;
	for (i=0; i<sizeOfVocabularyPointersList && charCount<sizeOfVocabularyBuffer; i++)
	{	vocabularyPointersList[i] = ptrIntoBuffer;
		if (i != sizeOfVocabularyPointersList-1)
			while ( *ptrIntoBuffer++ != '\0')
				charCount++;
		charCount++;
	}
	if (i < sizeOfVocabularyPointersList)
		cout << "We cannot find all " << sizeOfVocabularyPointersList << " words "
			 << "you want in the vocabulary buffer you specify." << endl;
	int numWordsFound = i;
	return numWordsFound;
}

/************************************************************************/
/************************************************************************/
void recognizeNoisyWordsInFileFromVocabulary
(char* noisyWordsFile, char* vocabularyFile, 
 char* ouputFile, int numOfLikelyWordsToSuggest, bool trace)
//Recongnize the noisyWrodToRecognize, suggest up to numOfLikelyWordsToSuggest 
//most likely words from the vocabularyFile, and write these words to the
//outputBuffer of size bufferSize.
//If trace is on, the noisy words and other information will also be record 
//into the outputFile.
{

	int i;
	//Read, store, and establish pointers to the words in the vocabulary
	ifstream vocabularyFileHandle;
	char vocabularyBuffer[VocabularyStorageLimit];
	int numWordsInVocabulary;
	numWordsInVocabulary = 
		readVocabularyFileIntoCharBuffer(vocabularyBuffer, VocabularyStorageLimit-1, vocabularyFile);

	char * * vocabularyPointersList;
	vocabularyPointersList = new char * [numWordsInVocabulary];
	getWordsPointersListFromCharBuffer
		(vocabularyPointersList, numWordsInVocabulary,
		 vocabularyBuffer, VocabularyStorageLimit);

	//Read, store, and establish pointers to noisy words
	ifstream noisyWordsFileHandle;
	char noisyWordsBuffer[NoisyWordsStorageLimit];
	int numNoisyWords;
	numNoisyWords = 
		readVocabularyFileIntoCharBuffer
		(noisyWordsBuffer, NoisyWordsStorageLimit-1, noisyWordsFile);

	char * * noisyWordsPointersList;
	noisyWordsPointersList = new char * [numNoisyWords];
	getWordsPointersListFromCharBuffer
		(noisyWordsPointersList, numNoisyWords,
		 noisyWordsBuffer, NoisyWordsStorageLimit);


	ofstream outputFileHandle;
	outputFileHandle.open(ouputFile);

	for (int i=0; i<numNoisyWords; i++)
	{	char outputBuffer[1000];
		recognize1NoisyWordFromVocabulary
		(	noisyWordsPointersList[i], 
			vocabularyPointersList, 
			numWordsInVocabulary,
			outputBuffer, 999,
			numOfLikelyWordsToSuggest
		);
		outputFileHandle << outputBuffer << endl;
		if (trace)
		 cout<< "THE WORD =" <<noisyWordsPointersList[i]<< endl
			 << "The most likely " << numOfLikelyWordsToSuggest 
			 << " words in the vocabulary are:" << endl
			 << outputBuffer << endl << endl;
	}
 outputFileHandle.close();
 delete[] vocabularyPointersList;
 delete[] noisyWordsPointersList;
}

